/**************************************************************
	bman_parse -- file table parser for BOMBERMAN64
	Decompression is LZSS 1K variant
***************************************************************
	1/7/2018 AC67
**************************************************************/

/***Decompressor***/
#define N		 1024	
#define F		   16	
#define THRESHOLD	2   	  
#define NIL			N	
/******************/

#define T_NUM 		14

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <ctype.h>

#include "bman_parse.h"

typedef unsigned long 	u64;
typedef unsigned int 	u32;
typedef unsigned short 	u16;
typedef unsigned char 	u8;

void decode(u8 *data, u32 compressedSize, u8 *output, u32 outSize, u32 pos);
void lzss_dec(u8* data, u32 compressedSize, u8* output, u32 outSize, u32 pos, u8 mode);

const u8 *rompath = "C:\\Users\\Rudje\\Desktop\\file table parser gcc\\Bomberman 64 (USA).z64";
char 	 *rom_buffer;
u8 		 t_num;

const int ftable_arr[T_NUM] =
{
		0x00120000, 0x00140000, 0x00160000, 
		0x00180000, 0x001C0000, 0x001E0000, 
		0x00200000, 0x00240000, 0x00260000,
		0x00280000, 0x002A0000, 0x002C0000,
		0X002E0000, 0x00300000
};

typedef struct {
	u32 table_entry;		/*address*/
	u32 offset_start_data;
	u32  table_index;		/*position in table header. initial 0x10 advance, with 0x8 increments*/
	u16 table_entry_count;	/*how many, always located 0x2008 ahead*/
	u32	table_offset;		/*offset from data start location address*/
	u32 fentry_csize;	/*compressed size address*/	
	u32 fentry_ucsize;	/*uncompressed size*/
	u8 *fentry_data;		/*compressed data*/
} ftable_t;

ftable_t tables[T_NUM];

u32 read4bytes(u8 *buf, u32 index)
{
	return ((u32)(buf[index+0] << 24 | 
				  buf[index+1] << 16 | 
				  buf[index+2] << 8  | 
				  buf[index+3]));
}

void parse_table(ftable_t t, u32 num)
{
	t.table_entry_count = (num == 13) ? 0x0368 : rom_buffer[t.table_entry + t.offset_start_data]; //table 13 needs manual set of entries, as cannot locate
	if(num==0xD) t.table_entry_count = 400;
	int temp = t.table_entry_count;
	t.table_index = 0x10;
	for(int i =0;i<t.table_entry_count;i++)
	{
		if(num == 0xD && i == 47)
			break;
		

		t.table_offset = read4bytes(rom_buffer, t.table_entry + t.table_index);
		if(t.table_offset == 0xFFFFFFFF) {
			printf("Error trying to read another entry (%08X)", t.table_offset);
			break;
		}
		

		t.fentry_csize = read4bytes(rom_buffer, (t.table_entry + t.table_index) + 0x4);
	
		if(t.fentry_csize > 0x00010000) { 
			printf("Next entry is huge. Not compressed file. Advancing...\n"); 
		} else {	
		t.fentry_ucsize = read4bytes(rom_buffer, (t.table_entry+t.offset_start_data)+t.table_offset);
		t.fentry_data = (u8*)malloc(sizeof(u8)+t.fentry_csize);
		for(u32 pos = 0; pos < t.fentry_csize; pos++) 
			t.fentry_data[pos] = rom_buffer[((t.table_entry+t.offset_start_data)+t.table_offset) + pos];
		
		/*debug data*/
		printf(						
		"[TABLE %02X %08X]\nTable Offset\t%08X\nCompressed Size\t%08X\nDecompressed Size\t%08X\nTable Index\t%04X\nEntry Count\t%02X [%i]\n\n",
		num, t.table_entry,
		t.table_offset, 
		t.fentry_csize, 
		t.fentry_ucsize, 
		t.table_index, 
		i, t.table_entry_count);

		char tmp[30];
		
		sprintf(tmp, "compressed//FT_%02X_F_%02X.bin", num, i);		
		FILE *d = fopen(tmp, "wb");
		fwrite(t.fentry_data, 1, t.fentry_csize, d);
		fclose(d);
	
		sprintf(tmp, "decompressed//FT_%02X_F_%02X.bin", num, i);		
		
		char tmp_decomp[t.fentry_ucsize];
		decode(t.fentry_data, t.fentry_csize, tmp_decomp, t.fentry_ucsize, 0x4);

		FILE *r = fopen(tmp, "wb");
		fwrite(tmp_decomp, 1, t.fentry_ucsize, r);
		fclose(r);

		}
		t.table_index += 0x8;
	}
}

void init_table_data()
{
	for(int i =0;i<T_NUM;i++)
	{
		tables[i].table_entry = ftable_arr[i];
		tables[i].offset_start_data = 0x2008;
		tables[i].table_index = 0;
		tables[i].table_entry_count = 0;
		tables[i].table_offset = 0;
		tables[i].fentry_csize = 0;
		tables[i].fentry_ucsize = 0;
	}
}

void read_rom()
{
	FILE *r = fopen(rompath, "rb"); 
		if(!r) puts("Error opening input file.");
	fseek(r, 0x0, SEEK_END);
	int stmp = ftell(r);
	fseek(r, 0x0, SEEK_SET);
	rom_buffer = malloc(sizeof(char)+stmp);
	if(fread(rom_buffer, 1, stmp, r)==0) 
		puts("Error reading input file.");
	fclose(r);
}

int main(int argc, char *argv[])
{

	printf("BOMBERMAN64 : RIPPING ALL FILE TABLES");
	read_rom();
	init_table_data();
	

		
	for(int x = 0;x<T_NUM;x++)
	{
		printf("Parsing Table %i\n", x);
		parse_table(tables[x], x);
	}
	
	return 0;
}

/*LZSS_0 1K Decoder ported from Zoinkity's Python Module*/
unsigned char fbuffer[N + F - 1];
void decode(u8* data, u32 compressedSize, u8* output, u32 outSize, u32 pos)	
{
	u32 flag, buf_l, inpos, outpos;
	u64 val, size, back;
	
	buf_l 	= 0;
	flag 	= 0;
	inpos 	= pos;
	val 	= 0;
	size 	= 0;
	outpos 	= 0;
	back 	= 0;
	
	for (u32 i = 0; i < 0x400; i++) //0x400 = 1kb
		fbuffer[i] = 0x00;

	buf_l = 0x3BE;
	
	while (inpos <= compressedSize)
	{
		flag >>= 1;
		if (flag < 2) 
		{
			flag = (data[inpos++] & 0xFF) | 0x100;
			if (flag < 0)
				break;
		}
		if (flag & 1)
		{
			val = data[inpos++] & 0xFF;
			if (val < 0)
				break;
			output[outpos++] = val;
			fbuffer[buf_l] = val;
			buf_l += 1;
			buf_l &= 0x3FF;
		}
		else 
		{
			back = data[inpos++];
			size = data[inpos++];

			back |= ((size << 2) & 0x300);
			size &= 0x3F;
			
			for (int i = 0; i < size+0x3; i++)
			{
				val = fbuffer[back];
				output[outpos++] = val;
				fbuffer[buf_l] = val;

				back += 1;
				back &= 0x3FF;

				buf_l += 1;
				buf_l &= 0x3FF;
			}
		}
	}
	
}

void lzss_dec(u8* data, u32 compressedSize, u8* output, u32 outSize, u32 pos, u8 mode)	
{
/*        """
    Three most-prevalent LZ-derived compression schemes.
    Can except modes:
        MIDWAY does not use a ring buffer
        LZSS for traditional LZSS with a ring buffer
        LZSS_0 for variation of LZSS with 00 NULLs o.0
        WILLIAMS for alternate decoding method
    Difference lies primarily in initialization buffer and values.
    LZSS is traditional lzss with 20 as fill character.
    LZSS_0 is not traditional; 0 is used as the NULL character.
    """*/
	int inputPosition = pos;
	int outputPosition = 0;

	unsigned char buffer[0x1000];

	int b_idx = 0;

    //# set up ring buffers for variations requiring it
    if (mode == 1)
	{
        for (int x = 0; x < 0x1000; x++)
			buffer[x] = 0;
        b_idx = 1;
	}
    else if ((mode == 2) || (mode == 3))
	{
        b_idx = 0xFEE;

		int bCounter = 0;

        for (int x = 0; x < b_idx; x++)
		{
			if (mode != 2)
				buffer[bCounter++] = 0;
			else
				buffer[bCounter++] = 32;
		}

        for (int x = 0; x < (0x1000-b_idx); x++)
			buffer[bCounter++] = 0;
	}
    unsigned long bits = 0;

     while (inputPosition < compressedSize)
	 {
        if (bits<0x100)
		{
            bits = data[inputPosition] | 0xFF00;
            inputPosition+=1;
		}
        if(bits&1)
		{
            if (mode!=1)
			{
                buffer[b_idx]=data[inputPosition];
                b_idx=(b_idx+1)&0xFFF;
			}
            output[outputPosition++] = data[inputPosition];
            inputPosition += 1;
		}
        else
		{
			unsigned long val;
			unsigned long size;

            //# size = data[inputPosition + (mode=="LZSS")]
            //# val = data[inputPosition + (mode!="LZSS")]
            if ((mode == 2) || (mode == 3))
			{
                val = data[inputPosition];
                size = data[inputPosition+1];
			}
            else
			{
                size = data[inputPosition];
                val = data[inputPosition+1];
                if ((size+val)==0)
					break;
			}
            inputPosition+=2;

            unsigned long back = size&0xF0;
            back*=0x10;
            back|= val;
            back&=0xFFF;
            size = (size&0xF) + 2 + (int)((mode == 2) || (mode == 3));
            for (int x = 0; x < size; x++)
			{
                if (mode != 1)
				{
                    buffer[b_idx] = buffer[back];
                    output[outputPosition++] = (buffer[back]);
                    b_idx=(b_idx+1)&0xFFF;
                    back = (back+1)&0xFFF;
				}
                else
				{
                    output[outputPosition++] = (output[outputPosition - back]);
				}
			}
		}
        bits>>=1;
	   }

}

